from pankmer.index import print_err, kmer_size, generate_mem_blocks, measure_genomes
from os.path import commonprefix, basename
from itertools import chain
from more_itertools import batched

from pankmer.version import __version__
from pankmer.parse_genomes_input import parse_genomes_input, remove_seq_ext

def dryrun(genomes_input=None, genomes_input_paired=None, split_memory: int = 1, threads: int = 1) -> dict:
    """Perform a dry run of indexing and return the metadata

    Parameters
    ----------
    genomes_input
        paths to input genomes, directories, or tar file
    genomes_input
        paths to input genomes or directory (paired)
    split_memory : int
        number of memory blocks per thread
    threads : int
        number of threads

    Returns
    -------
    dict
        index metadata
    """
    
    if genomes_input is None and genomes_input_paired is None:
        raise RuntimeError('genomes must be provided')
    print_err('Recording genome sizes')
    if genomes_input is not None:
        genomes_input, genomes, input_is_tar = parse_genomes_input(genomes_input)
        genomes_dict = measure_genomes(genomes, str(genomes_input[0]) if input_is_tar else '')
    else:
        genomes_dict, genomes, input_is_tar = {}, [], False
    if genomes_input_paired is not None:
        genomes_input_paired, genomes_paired, _= parse_genomes_input(genomes_input_paired)
        genomes_dict_paired = {commonprefix([g0, g1]): s0+s1
                               for ((g0, s0), (g1, s1))
                               in batched(sorted(measure_genomes(genomes_paired, '').items()), 2)}
    else:
        genomes_dict_paired, genomes_paired = {}, []
    mem_blocks = generate_mem_blocks(split_memory, threads)
    all_core_blocks = []
    for m in range(1, len(mem_blocks)):
        temp_core_block = [mem_blocks[m-1][-1]]
        for i in range(len(mem_blocks[m])):
            temp_core_block.append(mem_blocks[m][i])
            all_core_blocks.append(temp_core_block)
            temp_core_block = [mem_blocks[m][i]]
    return {'kmer_size': kmer_size(),
        'version': __version__,
        'genomes': {c: remove_seq_ext(basename(g))
                    for c, g in enumerate(chain(genomes,
                        (commonprefix([g0, g1])
                         for g0, g1 in batched(genomes_paired, 2))))},
        'genome_sizes': {remove_seq_ext(basename(g)): s
                         for g, s in chain(genomes_dict.items(),
                                           genomes_dict_paired.items())},
        'positions': {},
        'mem_blocks': all_core_blocks
    }



#===============================================================================
# collect.py
#===============================================================================

"""Calculate and plot collection curves"""

# Imports ======================================================================

import os.path
import pandas as pd
import seaborn as sns
from collections import Counter
import numpy as np
from math import prod, floor
from itertools import cycle
from scipy.stats import hypergeom



# Constants ====================================================================

COL_COLOR_PALETTE = 'mako_r'




# Functions ====================================================================

def col_values_conf(pk_results):
    pkr = pk_results
    sc_matrix = np.zeros((pkr.number_of_genomes,pkr.number_of_genomes))
    for _, score in pkr:
        s = pkr.decode_score(score)
        sc_matrix[sum(s) - 1,:] += np.array(s[::-1])
    sc_df = pd.DataFrame(sc_matrix, index=range(1, pkr.number_of_genomes+1),
        columns=tuple(str(os.path.basename(g)).replace('.fasta.gz', '').replace('.fa.gz', '')
                      for g in reversed(pkr.genomes)))
    g = len(sc_df.columns)
    coef_matrix_total = np.array([[prod((g-s-i)/(g-1-i) for i in range(n)) for n in range(g)]
                            for s in range(1,g+1)])
    coef_matrix_core = np.flip(coef_matrix_total.copy(), axis=0)
    for n in range(1,g):
        coef_matrix_total[:,n] += coef_matrix_total[:,n-1]
    col_df_core = pd.DataFrame(sc_df.transpose().dot(coef_matrix_core).transpose())
    col_df_total = pd.DataFrame(sc_df.transpose().dot(coef_matrix_total).transpose())
    col_df = pd.concat((col_df_total, col_df_core))
    col_df['n_genomes']= tuple(range(1,g+1))*2
    plotting_data = col_df.melt(
        id_vars=['n_genomes'], value_name='n_kmers')
    del plotting_data['variable']
    plotting_data['sequence'] = (('total',) * g + ('core',) * g) * g
    return plotting_data


def col_values(pk_results, contours=None):
    """Calculate collection curve

    Parameters
    ----------
    pk_results
        a PKResults object
    contours
        if not None, an iterable of integers between 0 and 100

    Yields
    -------
        collection curve values
    """

    pkr = pk_results
    g = pkr.number_of_genomes
    score_dist = Counter(sum(pkr.decode_score(score)) for _, score in pkr)
    for n_genomes in range(1, g+1):
        yield n_genomes, sum((1-prod((g-s-n)/(g-n) for n in range(n_genomes)))*score_dist[s]
            for s in range(1, g+1)), 'total'
        if contours:
            for c in contours:
                yield n_genomes, sum((hypergeom.sf(floor(c/100*n_genomes), g, s, n_genomes))*score_dist[s]
                    for s in range(1, g+1)), f'{c}%'
        yield n_genomes, sum(prod((s-n)/(g-n) for n in range(n_genomes))*score_dist[s]
            for s in range(1, g+1)), 'core'


def col_plot(plotting_data, output, title: str = 'Collection curve',
             linewidth: int = 3, palette=COL_COLOR_PALETTE, alpha: float = 1.0,
             width: float = 4.0, height: float = 3.0, legend_loc='best'):
    """Draw a plot of the collection curve

    Parameters
    ----------
    plotting_data : DataFrame
        data frame of plotting data to be passed to sns.lineplot
    output
        path to output file
    title : str
        plot title [Collection curve]
    linewidth : int
        line width [3]
    palette
        argument sent to seaborn to be used as color palette [mako_r]
    alpha : float
        opacity of plot lines [1.0]
    width : float
        width of plot in inches [4.0]
    height : float
        height of plot in inches [3.0]
    legend_loc : str
        location of plot legend, e.g. 'upper left', 'best', or 'outside' [best]
    """

    ax = sns.lineplot(x='n_genomes', y='n_kmers', hue='sequence',
                      data=plotting_data, linewidth=linewidth,
                      palette=palette, alpha=alpha)
    ax.set_title(title)
    ax.set_ylim(bottom=0)
    if legend_loc == 'outside':
        leg = ax.legend(bbox_to_anchor=(1.02, 1), loc='upper left',
                        borderaxespad=0)
    else:
        leg = ax.legend(loc=legend_loc)
    for line in leg.get_lines():
        line.set_linewidth(linewidth)
        line.set_alpha(alpha)
    fig = ax.get_figure()
    fig.set_figwidth(width)
    fig.set_figheight(height)
    fig.tight_layout()
    fig.savefig(output)
    fig.clf()


def collect(pk_results, output=None, title: str = 'Collection curve', linewidth: int = 3,
               palette=COL_COLOR_PALETTE, alpha=1, width=4, height=3,
               conf=False, contours=False, legend_loc='best'):
    """Compute a collection curve, optionally drawing intermediate contours,
    or generating a plot, or computing confidence intervals (experimental)

    Parameters
    ----------
    plotting_data : DataFrame
        data frame of plotting data to be passed to sns.lineplot
    output
        path to output file
    title : str
        plot title [Collection curve]
    linewidth : int
        line width [3]
    palette
        argument sent to seaborn to be used as color palette [mako_r]
    alpha : float
        opacity of plot lines [1.0]
    width : float
        width of plot in inches [4.0]
    height : float
        height of plot in inches [3.0]
    conf : bool
        draw experimental confidence intervals
    contours : bool or list of int
        if not false, should be a list of intermediate contours in percent
        (e.g. [25, 50, 75]) 
    legend_loc : str
        location of plot legend, e.g. 'upper left', 'best', or 'outside' [best]
    """

    if conf:
        col_df = col_values_conf(pk_results)
    else:
        col_df = pd.DataFrame(col_values(pk_results, contours=contours),
                              columns=('n_genomes', 'n_kmers', 'sequence'))
    if output:
        col_plot(col_df, output, title=title, linewidth=linewidth,
                 palette=(palette if contours else sns.color_palette(palette, n_colors=2)), alpha=alpha,
                 width=width, height=height, legend_loc=legend_loc)
    return col_df



import pandas as pd
import upsetplot
import gzip
from collections import Counter
from itertools import compress
from matplotlib import pyplot

def count_scores(pk_results, genomes, exclusive = False):
    """Compute score counts for input to upsetplot

    Parameters
    ----------
    pk_results : PKResults
        PKResults object representing superset index
    genomes
        iterable of genomes included in subset
    """

    genomes_ordered = tuple(g for g in pk_results.genomes if g in genomes)
    genomes_set = set(genomes)
    counts = {}
    for b, n in Counter(score for _, score in pk_results).items():
        expanded_score = pk_results.decode_score(b)
        if exclusive and bool(set(compress(pk_results.genomes, expanded_score)) - genomes_set):
            continue
        membership = tuple(sorted(set(compress(pk_results.genomes, expanded_score)).intersection(genomes_set)))
        if membership:
            counts[membership] = counts.get(membership, 0) + n
    return pd.Series(counts.values(), index=pd.MultiIndex.from_tuples(
        (tuple(g in k for g in genomes_ordered) for k in counts.keys()),
        names=genomes_ordered))


def upset(pk_results, output, genomes, vertical=False, show_counts=False,
          min_subset_size=None, max_subset_size=None, exclusive=False,
          table=None):
    score_counts = count_scores(pk_results, genomes, exclusive=exclusive)
    if table:
        with (gzip.open if table.endswith('.gz') else open)(table, 'wb') as f:
            score_counts.to_csv(f, sep='\t', header=['k-mers'])
    upsetplot.plot(score_counts,
        orientation='vertical' if vertical else 'horizontal',
        show_counts=show_counts,
        min_subset_size=min_subset_size,
        max_subset_size=max_subset_size)
    pyplot.savefig(output)



from pankmer.version import __version__
from pankmer.pankmer import PKResults, index_wrapper
from pankmer.collect import collect
from pankmer.adjacency_matrix import get_adjacency_matrix
from pankmer.clustermap import clustermap
from pankmer.anchor import anchor_region, anchor_genome
from pankmer.index import kmer_size
K = kmer_size()



import re
import os
import os.path

COORD_REGEX = re.compile(os.environ.get('PANKMER_COORD_REGEX', # Not used yet, but will be useful later
                                        '[\s\S]+:[0-9]+-[0-9]+$'))
                                        # '([Cc]hr)?[0-9XY]+:[0-9]+-[0-9]+$'))
GENES_PATH = os.environ.get('PANKMER_GENES_PATH') # Not used yet, but will be useful later

EXAMPLE_DATA_URL = 'https://salk-tm-pub.s3.us-west-2.amazonaws.com/PanKmer_example/PanKmer_example_Sp_Chr19.tar.gz'
EXAMPLE_DATA_DIR = os.environ.get('PANKMER_EXAMPLE_DATA_DIR', os.path.dirname(__file__))



import tarfile
import shutil
from pankmer.index import subset as _subset
from pankmer.version import __version__

def subset(pk_results, output, genomes, exclusive=False, gzip_level=6):
    output_is_tar = output.endswith('.tar')
    output_dir = output[:-4] if output_is_tar else output
    _subset(str(pk_results.results_dir),
                (str(pk_results.results_dir) if pk_results.input_is_tar else ''),
                 genomes, output, gzip_level, exclusive)
    if output_is_tar:
        with tarfile.open(output, 'w') as tar:
            tar.add(output_dir)
        shutil.rmtree(output_dir)



from Bio.bgzf import BgzfWriter
import gff2bed
import json
import math
import os.path
import os
import pybedtools
import pysam
import subprocess

from pankmer.tree import adj_to_jaccard
from pankmer.adjacency_matrix import get_adjacency_matrix
from pankmer.anchor import check_for_bgzip, get_chromosome_sizes_from_ref

LOWRES_STEP=100


def anchormap(pk_results, output_dir, anchors, anno, threads: int = 1):
    """Generate an anchormap for visualization

    Parameters
    ----------
    pk_results : PKResults
        a PKResults object
    output_dir
        directory for output files
    anchors
        iterable of paths to anchor genomes
    anno
        iterable of paths to annotation files
    threads : int
        number of threads to use
    """
    for anchor in anchors:
        check_for_bgzip(anchor)
    os.mkdir(output_dir)
    os.mkdir(os.path.join(output_dir, 'anchors'))
    os.mkdir(os.path.join(output_dir, 'anno'))
    print('loading index')
    pk_results.threads = threads
    print('writing distance matrix')
    (1 - adj_to_jaccard(get_adjacency_matrix(pk_results))).to_csv(
        os.path.join(output_dir, 'distance.tsv'), sep='\t')
    score_bitsize = math.ceil(pk_results.number_of_genomes/8)
    print('writing config')
    sizes = {'.'.join(os.path.basename(a).split('.')[:-2]):
             get_chromosome_sizes_from_ref(a).set_index('name')['size']
             for a in anchors}
    with open(os.path.join(output_dir,'config.json'), 'w') as f:
        json.dump({
            'prefix': output_dir,
            'genomes': list(pk_results.genomes),
            'anchors': ['.'.join(os.path.basename(a).split('.')[:-2])
                        for a in anchors],
            'lowres_step': LOWRES_STEP,
            'sizes': {a: [[chrom, size] for chrom, size in zip(s.index, s)]
                      for a, s in sizes.items()}},
            f)
    for anchor, anno in zip(anchors, anno):
        print('collecting regional scores')
        a = '.'.join(os.path.basename(anchor).split('.')[:-2])
        regcov_dict = pk_results.get_collapsed_regional_scores(anchor,
            {c: [[1, sizes[a][c]]] for c in sizes[a].index})
        print('converting to bytes')
        scores = b''.join(int(''.join(str(i) for i in s), 2).to_bytes(
                score_bitsize, byteorder="big", signed=False)
            for c in sizes[a].index for s in regcov_dict[c][(1, sizes[a][c])])
        print('writing scoremap')
        for step in 1, LOWRES_STEP:
            with BgzfWriter(os.path.join(output_dir, 'anchors', f'{a}.{step}.bgz'), 'wb') as scores_out:
                scores_out.write(scores[::step] if step > 1 else scores)
            subprocess.run(('bgzip', '-rI',
                os.path.join(output_dir, 'anchors', f'{a}.{step}.gzi'),
                os.path.join(output_dir, 'anchors', f'{a}.{step}.bgz')))
        print('formatting annotations')
        for type in 'gene', 'anno':
            bed = os.path.join(output_dir, 'anno', f'{a}.{type}.bed')
            tbx = os.path.join(output_dir, 'anno', f'{a}.{type}.bed.bgz')
            pybedtools.BedTool(gff2bed.convert(gff2bed.parse(anno))).saveas(bed).sort().saveas(bed)
            pysam.tabix_compress(bed, tbx, True)
            pysam.tabix_index(tbx, True, 0,1,2, csi=True)



import gzip
from io import BytesIO, TextIOWrapper

def nothing(x):
    return x

def gzip_agnostic_open(file_path, mode='rb', tar=None, **kwargs):
    """
    Open a file (e.g. FASTA) which may or may not be inside a tar file and
    may or may not be gzipped

    Parameters
    ----------
    file_path
        path to file, possibly relative to tar file
    mode
        file open mode, should be 'rb' or 'rt' [rb]
    tar
        tar file object containing file_path, or None
    **kwargs
        other arguments passed to open() or BytesIO()
    """
    
    return (gzip.open(tar.extractfile(file_path) if tar is not None
                      else file_path, mode, **kwargs)
            if str(file_path).endswith('.gz')
            else ((lambda b: TextIOWrapper(b) if mode[1]=='t' else b)(
                   BytesIO(tar.extractfile(file_path).read()), **kwargs)
                  if tar is not None else open(file_path, mode, **kwargs)))


import itertools
import tarfile
from os import listdir
from os.path import isfile, join, isdir, exists

SEQ_EXTENSIONS = (".fasta.gz", ".fna.gz", ".fa.gz", ".fastq.gz", ".fq.gz",
                ".fasta", ".fna", ".fa", ".fastq", ".fq")

def parse_genomes_input(genomes_input):
    if not isinstance(genomes_input, (tuple, list)):
        genomes_input = list(genomes_input) if isinstance(genomes_input, set) else [genomes_input]
    if len(genomes_input) == 1 and isfile(genomes_input[0]) and tarfile.is_tarfile(genomes_input[0]):
        input_is_tar = True
        with tarfile.open(genomes_input[0]) as tar:
            genomes = [tarinfo.name for tarinfo in tar if tarinfo.isreg()]
    else:
        input_is_tar = False
        genomes = list(itertools.chain.from_iterable(
            ([join(g, f) for f in sorted(listdir(g))] if isdir(g) else [g])
            for g in genomes_input))
        # Check if input files exist and are files
        for genome in genomes:
            if not exists(genome) or not isfile(genome):
                raise RuntimeError(f"{genome} does not exist or is not a file!")
    return genomes_input, genomes, input_is_tar

def remove_seq_ext(genome):
    for ext in SEQ_EXTENSIONS:
        if genome.endswith(ext):
            genome = genome[:-1*len(ext)]
    return genome


import urllib3
import shutil
import os
import os.path
import tarfile
import ftplib
import tempfile
import subprocess
from pankmer.env import EXAMPLE_DATA_URL, EXAMPLE_DATA_DIR
from pankmer.datasets import (SPOLYRHIZA_URL, SPOLYRHIZA_FILES,
                              SLYCOPERSICUM_URL_FORMAT, SLYCOPERSICUM_IDS,
                              ZMAYS_URL, ZMAYS_FILES, HSAPIENS_URL, HSAPIENS_AGC,
                              HSAPIENS_AGC_BINARY_URL,
                              HSAPIENS_IDS, ATHALIANA_URL_FORMAT, ATHALIANA_IDS,
                              BSUBTILIS_DATA_FTP, BSUBTILIS_DATA_PATHS)

def download_example(dir: str = EXAMPLE_DATA_DIR, species=None,
                     n_samples: int = 1):
    """Download an example datatset

    Parameters
    ----------
    dir : str
        Destination directory for example data
    species
        If None, download small example dataset. if str, download publicly
        available genomes. Species: max_samples. Spolyrhiza: 3, Bsubtilis: 164,
    n_samples : int
        Number of samples to download, must be less than max [1]
    """

    if species == 'Spolyrhiza':
        if n_samples > 3:
            raise RuntimeError('n_samples parameter must be <= 3 for Spolyrhiza')
        http = urllib3.PoolManager()
        with tempfile.TemporaryDirectory(dir=dir) as temp_dir:
            tar_file_path = os.path.join(temp_dir, os.path.basename(SPOLYRHIZA_URL))
            if os.path.exists(tar_file_path[:-7]):
                raise RuntimeError('destination already exists')
            with http.request('GET', SPOLYRHIZA_URL, preload_content=False) as r, open(tar_file_path, 'wb') as out_file:
                shutil.copyfileobj(r, out_file)
            with tarfile.open(tar_file_path) as tar:
                tar.extractall(temp_dir)
            os.remove(tar_file_path)
            for fasta_file in SPOLYRHIZA_FILES[:n_samples]:
                shutil.move(os.path.join(temp_dir, 'Spolyrhiza_genomes', fasta_file),
                                os.path.join(dir, fasta_file))
            shutil.rmtree(os.path.join(temp_dir, 'Spolyrhiza_genomes'))
    
    elif species == 'Slycopersicum':
        if n_samples > 46:
            raise RuntimeError('n_samples parameter must be <= 46 for Slycopersicum')
        http = urllib3.PoolManager()
        for genome_id in SLYCOPERSICUM_IDS[:n_samples]:
            fasta_url = SLYCOPERSICUM_URL_FORMAT.format(genome_id)
            dest_file_path = os.path.join(dir, os.path.basename(fasta_url))
            if os.path.exists(dest_file_path):
                raise RuntimeError('destination already exists')
            with http.request('GET', fasta_url, preload_content=False) as r, open(dest_file_path, 'wb') as out_file:
                shutil.copyfileobj(r, out_file)
    
    elif species == 'Zmays':
        if n_samples > 54:
            raise RuntimeError('n_samples parameter must be <= 54 for Zmays')
        http = urllib3.PoolManager()
        for fasta_file in ZMAYS_FILES[:n_samples]:
            fasta_url = os.path.join(ZMAYS_URL, fasta_file)
            dest_file_path = os.path.join(dir, os.path.basename(fasta_url))
            if os.path.exists(dest_file_path):
                raise RuntimeError('destination already exists')
            with http.request('GET', fasta_url, preload_content=False) as r, open(dest_file_path, 'wb') as out_file:
                shutil.copyfileobj(r, out_file)
    
    elif species == 'Hsapiens':
        if n_samples > 94:
            raise RuntimeError('n_samples parameter must be <= 94 for Hsapiens')
        http = urllib3.PoolManager()
        with tempfile.TemporaryDirectory(dir=dir) as temp_dir:
            tar_file_path = os.path.join(temp_dir, os.path.basename(HSAPIENS_AGC_BINARY_URL))
            agc_binary_path = os.path.join(temp_dir, 'agc-1.1_x64-linux', 'agc')
            agc_file_path = os.path.join(temp_dir, HSAPIENS_AGC)
            with http.request('GET', HSAPIENS_AGC_BINARY_URL, preload_content=False) as r, open(tar_file_path, 'wb') as out_file:
                shutil.copyfileobj(r, out_file)
            with tarfile.open(tar_file_path) as tar:
                tar.extractall(temp_dir)
            with http.request('GET', HSAPIENS_URL, preload_content=False) as r, open(agc_file_path, 'wb') as out_file:
                shutil.copyfileobj(r, out_file)
            for genome_id in HSAPIENS_IDS[:n_samples]:
                with open(os.path.join(dir, f'{genome_id}.fa'), 'wb') as f:
                    subprocess.run(
                        (agc_binary_path, 'getset', agc_file_path, genome_id),
                        stdout=f)
    
    elif species == 'Bsubtilis':
        if n_samples > 164:
            raise RuntimeError('n_samples parameter must be <= 164 for Bsubtillis')
        ftp = ftplib.FTP(BSUBTILIS_DATA_FTP)
        ftp.login()
        for ftp_path in BSUBTILIS_DATA_PATHS[:n_samples]:
            with open(os.path.join(dir, os.path.basename(ftp_path)), 'wb') as f:
                ftp.retrbinary(f'RETR {ftp_path}', f.write)

    elif species == 'Athaliana':
        if n_samples > 1135:
            raise RuntimeError('n_samples parameter must be <= 1135 for Athaliana')
        http = urllib3.PoolManager()
        for pseudo in ATHALIANA_IDS[:n_samples]:
            fasta_url = ATHALIANA_URL_FORMAT.format(pseudo)
            dest_file_path = os.path.join(dir, os.path.basename(fasta_url))
            if os.path.exists(dest_file_path):
                raise RuntimeError('destination already exists')
            with http.request('GET', fasta_url, preload_content=False) as r, open(dest_file_path, 'wb') as out_file:
                shutil.copyfileobj(r, out_file)

    elif isinstance(species, str):
        raise RuntimeError('invalid species')

    else:
        http = urllib3.PoolManager()
        tar_file_path = os.path.join(dir, os.path.basename(EXAMPLE_DATA_URL))
        if os.path.exists(tar_file_path[:-7]):
            raise RuntimeError('destination already exists')
        with http.request('GET', EXAMPLE_DATA_URL, preload_content=False) as r, open(tar_file_path, 'wb') as out_file:
            shutil.copyfileobj(r, out_file)
        with tarfile.open(tar_file_path) as tar:
            tar.extractall(dir)
        os.remove(tar_file_path)
